package beanstalkd

import (
	"flag"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/common/tcp"
)

var log = _log.New("beanstalkd")

var IGNORE_METRICS = []string{
	"pid",
	"version",
	"rusage-utime",
	"rusage-stime",
	"id",
	"hostname",
}

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "beanstalkd.timeout", confTimeout, "[beanstalkd] operation timeout")
	fs.DurationVar(&confInterval, "beanstalkd.interval", confInterval, "[beanstalkd] interval between polls")
}

func MakeDaemon(namespace string) *tcp.TcpDaemon {
	return &tcp.TcpDaemon{
		Timeout:        confTimeout,
		ConfInterval:   confInterval,
		Namespace:      namespace,
		ServiceType:    "beanstalkd",
		Keepalive:      true,
		CollectService: collect,
	}
}

func collect(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {
	stats("", d, svc, m, target)
	tubes := listTubes(d, svc, m, target)
	for _, tube := range tubes {
		stats(tube, d, svc, m, target)
	}
}

func stats(tube string, d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {
	var cmd string
	if tube == "" {
		cmd = "stats"
		tube = "*"
	} else {
		cmd = "stats-tube " + tube
	}
	data, err := d.Command(svc, cmd+"\r\n", []string{""})
	if err != nil {
		m.IncrError("beanstalkd", "beanstalkd")
		log.Warnln("error in collecting beanstalkd metrics", svc, err)
		return
	}

	for _, line := range data {
		kv := strings.SplitN(line, ": ", 2)
		if len(kv) != 2 {
			//TODO log
			continue
		}
		stat := kv[0]
		if common.Index(IGNORE_METRICS, stat) > -1 {
			continue
		}
		val, err := strconv.ParseFloat(kv[1], 64)
		if err != nil {
			//TODO log
			continue
		}
		labels := [][]string{
			[]string{"namespace", d.Namespace},
			[]string{"service", svc.ID},
			[]string{"service_type", svc.Type},
			[]string{"service_addr", target},
		}

		// current-jobs-urgent, current-jobs-ready, current-jobs-reserved,
		// current-jobs-delayed, current-jobs-buried
		prefixJobs := "current-jobs-"
		prefixCmd := "cmd-"
		if strings.HasPrefix(stat, prefixJobs) {
			labels = append(labels, []string{"status", stat[len(prefixJobs):]})
			labels = append(labels, []string{"tube", tube})
			stat = "jobs"

			// cmd-put, cmd-peek, cmd-peek-ready, cmd-peek-delayed
			// cmd-peek-buried, cmd-reserve, cmd-reserve-with-timeout
			// cmd-delete, cmd-release, cmd-use, cmd-watch, cmd-ignore
			// cmd-bury, cmd-kick, cmd-touch, cmd-stats, cmd-stats-job
			// cmd-stats-tube, cmd-list-tubes, cmd-list-tube-used
			// cmd-list-tubes-watched, cmd-pause-tube
		} else if strings.HasPrefix(stat, prefixCmd) {
			labels = append(labels, []string{"command", stat[len(prefixCmd):]})
			stat = "commands_total"

		} else {
			switch stat {

			case "job-timeouts":
				stat = "job_timeouts_total"

			case "total-jobs":
				stat = "jobs_total"
				labels = append(labels, []string{"tube", tube})

			case "max-job-size":
				stat = "job_size_bytes_max"

			case "current-tubes":
				stat = "tubes"

			case "current-connections":
				stat = "connections"
				labels = append(labels, []string{"type", "open"})
				labels = append(labels, []string{"tube", tube})
			case "current-producers":
				stat = "connections"
				labels = append(labels, []string{"type", "producer"})
				labels = append(labels, []string{"tube", tube})
			case "current-workers":
				stat = "connections"
				labels = append(labels, []string{"type", "worker"})
				labels = append(labels, []string{"tube", tube})
			case "current-waiting":
				stat = "connections"
				labels = append(labels, []string{"type", "waiting"})
				labels = append(labels, []string{"tube", tube})
			case "current-using":
				stat = "connections"
				labels = append(labels, []string{"type", "using"})
				labels = append(labels, []string{"tube", tube})
			case "current-watching":
				stat = "connections"
				labels = append(labels, []string{"type", "watching"})
				labels = append(labels, []string{"tube", tube})

			case "total-connections":
				stat = "connections_total"

			case "uptime":
				stat = "uptime_seconds"

			case "binlog-records-migrated":
				stat = "binlog_records"
				labels = append(labels, []string{"type", "migrated"})
			case "binlog-records-written":
				stat = "binlog_records"
				labels = append(labels, []string{"type", "written"})

			case "binlog-max-size":
				stat = "binlog_size_bytes_max"

			case "pause":
				stat = "tube_pause_seconds"
				labels = append(labels, []string{"tube", tube})
			case "pause-time-left":
				stat = "tube_pause_left_seconds"
				labels = append(labels, []string{"tube", tube})

			/* by default use the metrics as-is and consider it's a float */
			default:
				/* non exhuastiv list:
				 *   binlog-oldest-index
				 *   binlog-current-index
				 */
				stat = strings.ReplaceAll(stat, "-", "_")
			}
		}
		m.Set(&common.Metric{
			Name:   "beanstalkd_" + stat,
			Labels: labels,
			Value:  val,
		})
	}
}

func listTubes(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) []string {

	data, err := d.Command(svc, "list-tubes\r\n", []string{""})
	if err != nil {
		m.IncrError("beanstalkd", "beanstalkd")
		log.Warnln("error in listing beanstalkd tubes", svc, err)
		return nil
	}

	tubes := []string{}

	for _, line := range data {
		kv := strings.SplitN(line, " ", 2)
		if len(kv) != 2 || kv[0] != "-" || kv[1] == "" {
			continue
		}
		tubes = append(tubes, kv[1])
	}
	return tubes
}
